#ifndef IVEC_H
#define IVEC_H
#include <stdbool.h>
#include <stddef.h>
/* Struct Declaration. The actual struct definition is hidden in `ivec.c`.
 *
 * This prevents people, who are using our vector library, to directly access
 * the fields of this struct, like `capacity`. The only way to interact with the
 * vector is via the functions declared in this header.
 *
 * In other words: people using this library are not able to do stupid stuff
 * like changing the vector's capacity without also reallocating the memory area
 * of the vector.
 */
struct Ivec;

/* Allows us to write `Ivec` instead of `struct Ivec`. */
typedef struct Ivec Ivec;

/* Returns a pointer to a freshly allocated vector of length 0 and capacity 1.
 * Returns NULL if we run out of memory.
 */
Ivec *ivec_new();

/* Free the dynamically allocated memory areas of `xs`. */
void ivec_free(Ivec *xs);

/* Returns a pointer to the element at index `i` of vector `xs`. */
int *ivec_at(Ivec *xs, size_t i);

/* Returns the number of elements currently stored in `xs`. */
size_t ivec_length(Ivec *xs);

/* Returns the vector's current capacity. */
size_t ivec_capacity(Ivec *xs);

/* Append a single element `x` at the end of the vector `xs`.
 *
 * If the vector's length is equal to its capacity, the vector's data is
 * reallocated at the double of its capacity to allow for amortized constant
 * runtime.
 *
 * Returns `false` if we run out memory, otherwise `true`.
 */
bool ivec_push(Ivec *xs, int x);

/* Remove the last element of vector `xs` if it exists. Otherwise do nothing.
 *
 * If the resulting vector's length is half of its capacity, the vector's
 * capacity is shrunk to its new length to save memory.
 */
void ivec_pop(Ivec *xs);

/* Print the address, length, capacity, and elements of `xs`. */
void ivec_print(Ivec *xs);

/* Returns a pointer to the smallest element in `xs` which has an index greater
 * or equal to `begin` but less than `end`.
 *
 * If the range between `begin` and `end` is invalid or contains no elements,
 * then the NULL-pointer is returned.
 */
int *ivec_min_between(Ivec *xs, size_t begin, size_t end);

/* Sorts the elements of `xs` in increasing order. */
void ivec_sort(Ivec *xs);

/* Replace every element x of `xs` by the return value of `func(x)` */
void ivec_map(Ivec *xs, int (*func) (int));

/* Apply function `func` to every element of `xs` */
void ivec_iter(Ivec *xs, void (*func) (int*));

#endif /* IVEC_H */
