// Copyright 2019 University of Freiburg
// Author: Peter Thiemann <thiemann@informatik.uni-freiburg.de>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "./memtools.h"
#include "./json_data.h"
// Data Structures /////////////////////////////////////////////////////////////
struct JsonObject {
  Vec* members;
};

struct JsonMember {
  char* name;
  JsonValue* value;
};

struct JsonValue {
  JsonValueType type;
  union {
    double as_number;
    char* as_string;
    Vec* as_array;
    JsonObject* as_object;
  } value;
};

// JsonObject //////////////////////////////////////////////////////////////////
JsonObject* json_object_new(Vec* members) {
  JsonObject* o = malloc_or_exit(sizeof(JsonObject));
  o->members = members;
  return o;
}

Vec* json_object_members(JsonObject* o) {
  return o->members;
}

JsonValue* json_object_find(JsonObject* o, const char* name) {
  for (size_t i = 0; i < vec_length(o->members); ++i) {
    JsonMember* member = *vec_at(o->members, i);
    if (strcmp(name, member->name) == 0) {
      return member->value;
    }
  }
  return NULL;
}

void json_object_free(JsonObject* o) {
  for (size_t i = 0; i < vec_length(o->members); ++i) {
    JsonMember* member = *vec_at(o->members, i);
    json_member_free(member);
  }
  vec_free(o->members);
  free(o);
}

// JsonMember //////////////////////////////////////////////////////////////////
// assume that name and value are safe to share
JsonMember* json_member_new(char * name, JsonValue* value) {
  JsonMember* m = malloc_or_exit(sizeof(JsonMember));
  m->name = name;
  m->value = value;
  return m;
}

char* json_member_name(JsonMember* m) {
  return m->name;
}

JsonValue* json_member_value(JsonMember* m) {
  return m->value;
}

void json_member_free(JsonMember* m) {
  free(m->name);
  json_value_free(m->value);
  free(m);
}

// JsonValue ///////////////////////////////////////////////////////////////////
JsonValue* json_value_new_null() {
  JsonValue* v = malloc_or_exit(sizeof (JsonValue));
  v->type = JSON_NULL;
  return v;
}

JsonValue* json_value_new_true() {
  JsonValue* v = malloc_or_exit(sizeof (JsonValue));
  v->type = JSON_TRUE;
  return v;
}

JsonValue* json_value_new_false() {
  JsonValue* v = malloc_or_exit(sizeof (JsonValue));
  v->type = JSON_FALSE;
  return v;
}

JsonValue* json_value_new_array(Vec *vs) {
  JsonValue* v = malloc_or_exit (sizeof (JsonValue));
  v->type = JSON_ARRAY;
  v->value.as_array = vs;
  return v;
}

JsonValue* json_value_new_string(char* s) {
  JsonValue* v = malloc_or_exit(sizeof(JsonValue));
  v->type = JSON_STRING;
  v->value.as_string = s;
  return v;
}

JsonValue* json_value_new_number(double d) {
  JsonValue* v = malloc_or_exit(sizeof(JsonValue));
  v->type = JSON_NUMBER;
  v->value.as_number = d;
  return v;
}

JsonValue* json_value_new_object(JsonObject* o) {
  JsonValue* v = malloc_or_exit(sizeof(JsonValue));
  v->type = JSON_OBJECT;
  v->value.as_object = o;
  return v;
}

void json_value_free(JsonValue* v) {
  switch (v->type) {
  case JSON_NULL:
  case JSON_TRUE:
  case JSON_FALSE:
  case JSON_NUMBER:
    break;
  case JSON_STRING:
    free(v->value.as_string);
    break;
  case JSON_OBJECT:
    json_object_free(v->value.as_object);
    break;
  case JSON_ARRAY:
    {
      Vec *vs = v->value.as_array;
      for (size_t i=0; i < vec_length (vs); i++) {
	json_value_free (*vec_at (vs, i));
      }
    }
    break;
  }
  free(v);
}

JsonValueType json_value_type(JsonValue* v) {
  return v->type;
}

char* json_value_as_string(JsonValue* v) {
  return v->value.as_string;
}

double json_value_as_number(JsonValue* v) {
  return v->value.as_number;
}

JsonObject* json_value_as_object(JsonValue* v) {
  return v->value.as_object;
}

void json_value_print (JsonValue *v) {
  switch (v->type) {
  case JSON_NULL:
    printf ("null\n");
    break;
  case JSON_TRUE:
    printf ("true\n");
    break;
  case JSON_FALSE:
    printf ("false\n");
    break;
  case JSON_NUMBER:
    printf ("%g\n", v->value.as_number);
    break;
  case JSON_STRING:
    printf ("\"%s\"\n", v->value.as_string);
    break;
  case JSON_ARRAY:
    {
      Vec *vs = v->value.as_array;
      printf ("[ ");
      if (vec_length (vs) > 0) {
	json_value_print (*vec_at (vs, 0));
	for (size_t i=1; i < vec_length (vs); i++) {
	  printf (", ");
	  json_value_print (*vec_at (vs, i));
	}
      }
      printf ("]\n");
    }
    break;
  case JSON_OBJECT:
    { Vec *ms = json_object_members(v->value.as_object);
      printf ("{ ");
      if (vec_length (ms) > 0) {
	JsonMember *m = *vec_at (ms, 0);
	printf ("\"%s\" : ", m->name);
	json_value_print (m->value);
	for (size_t i=1; i < vec_length (ms); i++) {
	  m = *vec_at (ms, i);
	  printf (", \"%s\" : ", m->name);
	  json_value_print (m->value);
	}
      }
      printf ("}\n");
    }
    break;
  }
}

